{-
  Generate C function implementing img. sampling function

  See (or change) plugin_root below for location of temp files.  Make sure
  the directory exists.

-}
module CCompile where

import List ( findIndex, find, intersperse )
import Int ( fromInt )
import qualified Pretty as PP
import IOExts (trace,unsafePerformIO)
import NumExts (floatToDouble)
import Maybe ( isJust )

import Types
import qualified F32vec4 as F
import DExp
import PP
import Cse (rename)
import CompileUtils
import qualified Exp


type CCode = PPDoc

showCDoc :: CCode -> String
showCDoc x = showPPDoc x 0

compile :: Exp -> CgM CCode
compile (LitInt n)   = return (ppInt n)
compile (LitDouble x) = return (ppDouble x)
compile (LitBool  b) = return (ppInt (if b then 1 else 0))
compile (ImportData x) = do
  n <- newImport x
  return (uncurry ppImportVar n)
compile (Var id _)   = do
  locals <- getLocalEnv
  exts   <- getExterns
  params <- getParams
  x      <- 
   case findIndex ((== id) . fst) params of
     Just n   -> return (text id)
     Nothing  -> 
         case (lookup id exts) of
           Just (v,_) -> return (text v)
  	   _ ->
            case  find ((== id) . fst) locals of
              Just (_,(m,_)) -> return (ppVar m)
              Nothing    -> 
                  error ("Failed to find " ++ show id ++
                         ".  Let-bound: " ++
                           showSep showBoundLocal locals ++
                         "; params: " ++ showSep showParam params
			)
  if id == "X3" then
     {-trace (showPPDoc x 0 ++ show (locals,exts))-} (return x)
   else
     return x 

compile (Let name rhs body) = do
    -- Generate code for binding's RHS.
  flg     <- atStatementLevel
  loc     <- newLocal' name (typeOf rhs)
  assignC <- 
    if flg then do
       compileAssign (ppLocal loc) name rhs
      else do
       rhsC  <- asExpression (compile rhs)
       return (ppLocal loc <+> equals <+> rhsC )
   -- Compile the remaining bindings and body with a new bound local and a
   -- removed free local. 
  bodyC <- compile body
  if flg then
      return (assignC $$ bodyC)
   else do
      return (parens ( assignC <+> comma <+> bodyC))

compile (Bind (NewRef init) (Lambda x ty body)) = do
  initC        <- compile init
  (bodyC,loc) <- newLocal x ty (const (compile body))
  return (ppLocal loc <+> equals <+> initC <> semi $$ bodyC)

compile (Bind e@(NewArr sz _) (Lambda x ty body)) = do
  (bodyC, loc) <- newLocal x ty (const (compile body))
  arrC         <- compileAssign (ppLocal loc) x e
  return (arrC $$ bodyC)

compile (ReadRef ref@(Var refVar (Ref ty))) = do
  locals <- getLocalEnv
  let
     loc =
         case  find ((== refVar) . fst) locals of
           Just (_,l) -> l
  return (ppLocal loc <> semi)

compile (WriteRef ref@(Var refVar _) val) = do
  locals <- getLocalEnv
  let
     loc =
         case  find ((== refVar) . fst) locals of
           Just (_,l) -> l
  compileAssign (ppLocal loc) refVar val

{-
compile (UpdateRef ref@(Var refVar (Ref ty)) (Lambda x _ action)) = do
  refC    <- compile ref
--  loc     <- newLocal' refVar (Ref ty)
  locals <- getLocalEnv
  let
     loc =
         case  find ((== refVar) . fst) locals of
           Just (_,l) -> l
  code    <- compileAssign (ppLocal loc) refVar (rename x refVar action)
  return (ppLocal loc <+> equals <+> refC <> semi $$ code)
-}

compile (If condE thenE elseE) = do
  condC <- asExpression (compile condE)
  thenC <- compile thenE
  elseC <- compile elseE
  joinIf condC thenC elseC

compile (Loop varName limit body) = do
   limitC           <- compile limit
   (bodyC, iterLoc) <- arrayScope $ newLocal varName Int (const (compile body))
   let pp_local = ppLocal iterLoc
   return $
    text "for" <+> parens (pp_local <+> equals   <+> char '0' <+> semi <+>
   			   pp_local <+> char '<' <+> limitC   <+> semi <+>
			   pp_local <> text "++" ) <+> char '{' $$
    nest 3 bodyC $$
    char '}'

-- (conal) Does this case produce legit C code?  Doesn't look like it to me.
compile e@(MkArr size i elem) =
  error ("compile: orphaned MkArr\n" ++ show (Exp.mkE e)) $ do
    loc  <- newLocal' i (typeOf e)
    code <- compileAssign (ppLocal loc) i e
    return (code $$ ppLocal loc)

compile (ReadArr array index) = do
  arrayC <- compile array
  indexC <- compile index
  return $ arrayC <> text "[" <> indexC <> text "]"

compile (ReadMutArr array index) = do
  arrayC <- compile array
  indexC <- compile index
  return $ arrayC <> text "[" <> indexC <> text "]"

compile (WriteArr array index e@(Let{})) = 
  compile (floatBeyondLet (WriteArr array index) e)

compile (WriteArr array index val) =   do
  lhsC   <- compile (ReadArr array index)
  compileAssign lhsC "write-arr-name-not-used" val

compile (Add e (Negate e')) = compile (Sub e e')
compile (Add (Negate e') e) = compile (Sub e e')

-- No longer necessary, due to added unary negation in C++
-- compile (Negate e) | typeOf e == F32vec4 =
--   compile (Sub (zero F32vec4) e)

compile (Mul e (Recip e')) = compile (Divide e e')

-- F32vec4 has recip support
compile (Recip e) | typeOf e == Double = compile (Divide (LitDouble 1) e)

-- a * 2^k = a <<< k
compile (Mul (LitInt n) e) 
 | frac == 0 = compile (LShift e (LitInt k))
 where
   (k,frac) = properFraction (logBase 2 (fromInt n))

-- a `div` 2^k = a >>> k
compile (Div e (LitInt n))
 | frac == 0 = compile (RShift e (LitInt k))
 where
   (k,frac) = properFraction (logBase 2 (fromInt n))

-- a `mod` 2^k = a .&. (2^k - 1)
compile (Mod e (LitInt n))
 | frac == 0 = compile (BitAnd e (LitInt (n-1)))
 where
   (k,frac) = properFraction (logBase 2 (fromInt n))

compile p@(Pow b a@(LitDouble d)) | frac == 0 =
  if isSimple b then
     -- Expand to efficient binary tree of multiplies,
     -- possibly with an outer recip.  See Prelude.
     compile (b ^^ k)  
  else -- Make base simple and retry.
       asExpression $  -- awkward to require here
       compile $
       Let baseName b (Pow (Var baseName Double) a)
 where
   (k,frac) = properFraction d
   baseName = unsafePerformIO (newId' "base")

   -- Swiped from Prelude, but changed to use Mul instead "*".  That way,
   -- we don't let a*a get rewritten back into a**2.
   --- (^)            :: (Num a, Integral b) => a -> b -> a
   x ^ n  | n > 0  = f x (n-1) x
                     where f _ 0 y = y
                           f x n y = g x n where
                                     g x n | even n    = g (x`Mul`x) (n`quot`2)
                                           | otherwise = f x (n-1) (x`Mul`y)
   _ ^ _           = error "CCompile.^: bogus exponent"

   -- (^^)           :: (Fractional a, Integral b) => a -> b -> a
   x ^^ n          = if n >= 0 then x ^ n else Recip (x^(-n))


compile (Pow b (LitDouble 0.5)) = compile (Sqrt b)

-- Or maybe the following, if we get rounding errors in the exponent
-- 
-- compile (Pow b (LitDouble d)) | isHalf = compile (Sqrt b)
--  where
--   isHalf = abs (d - 0.5) < 0.001


compile Noop      = return (text "/* empty */" <> empty)
compile Rand      = return (text "rand();")

-- (conal) Is this transformation correct?
compile (LShift e (LitInt n)) 
  | n < 0 = error "compile: left shift ==> right shift\n" $
            compile (RShift e (LitInt (-n)))

compile (SetInt addr val) = do
  addrC  <- asExpression (compile addr)
  compileAssign (text "*(int*)" <> parens addrC) "unused-SetInt-name" val

{-
  For (Bind a (Lambda v _ b)), we need to bind the result of 
  performing 'a' to v (but only if v is actually used), so the
  assignment is pushed inwards using 'compileAssign'.
-}
compile (Bind act1 (Lambda v ty act2)) = do
  actC1 <- 
	if v `occFreeIn` act2 then do
  	   loc   <- newLocal' v ty
	   compileAssign (ppLocal loc) v act1
	 else
	   compile act1 		
  actC2 <- compile act2
  return (actC1 <> semi $$ actC2)

compile (Perform e) =
  error ("CCompile.compile: orphaned Perform:\n" ++ show (Exp.mkE e))

compile (Return e) = compile e

compile (DoProgress val max) = do
   valC <- compile val
   maxC <- compile max
   return (text "if (doProgress) { ((ProgressProc)pProc)" <> ppTuple [valC, maxC] <> semi <> text "}")

compile (Copy1DArray bptr len) = do
   bptrC <- compile bptr
   lenC  <- compile len
   return (text "CopyRow" <> ppTuple [bptrC, lenC] <> semi)

-- (conal) I think that ";" insertion for statements ought to be done at
-- each simple statement, e.g. for SetInt, rather than in compounds as
-- it's done now.  If so, remove both "<> semi" in clause above.  The way
-- it is now, "a >>* b >>* c" will turn into "a' ; b' ; c';;".

compile (NthF32 i v) = do
  vc <- compile v
  ic <- compile i
  -- ToDo: add parens around vc only when necessary.
  return $ parens vc <> text "[" <> ic <> text "]"

-- The four arguments to PackF32 are often equal, as generated by the
-- replF32 function (e.g., for scalar multiplication).  Specializing for
-- this case yields better asm code.  In a simple spinning imported
-- bitmap, we get about 10% speed-up.
compile (PackF32 e e2 e3 e4) | all (== e) [e2,e3,e4] = do
  vc <- compile e
  return $ text "F32vec4" <> parens vc

compile appl = do
  env <- getEnv
  asExpression $
   case arity of
     0 -> return (ppOp env op)
     1 -> do
      e1c <- compile e1
      return (ppOp env op <> setPrec precOp (wrap e1c))
     2 -> do
      e1c <- compile e1
      e2c <- compile e2
      let app = e1c <+> ppOp env op <+> (onRight e2c)
      return $
             if is_prim then
                setPrec precOp app
             else ppOp env op <> ppTuple [e1c,e2c]
     otherwise -> do
       ecs <- mapM compile args
       return $ ppOp env op <> ppTuple ecs
 where
   (op, args) = toHnf appl
   arity      = opArity op
   is_prim    = opIsPrim op
   
   precOp     = opPrec op
   
   setPrec Nothing  e = e
   setPrec (Just x) e = ppOperator (x, opAssoc op) e

   wrap d
    | is_prim   = d
    | otherwise = parens d

   (e1:es1) = args
   (e2:es2) = es1

   ppOp env op = text ((opMap env) op)

compile exp = error ("compile: I don't yet handle " ++ show exp)

{-
 Compiling the assignment of a temporary to a value - called when
 compiling a Let at a non-statement level.
-}
compileAssign :: CCode -> String -> Exp -> CgM CCode
compileAssign loc nm z@(Let i rhs e) = do
  nloc    <- newLocal' i (typeOf rhs)
  assignI <- compileAssign (ppLocal nloc) i rhs
  assignN <- compileAssign loc nm e
  return (assignI $$ assignN)    

compileAssign loc nm (If c e1 e2)  = do
  condC <- asExpression (compile c)
  thenC <- compileAssign loc nm e1
  elseC <- compileAssign loc nm e2
  joinIf condC thenC elseC

compileAssign loc name (MkArr size i elem) | isSimple size = do
  sizeC  <- compile size
  (initC, allocC)
      <- allocArray loc size sizeC elemTy $ 
         asStatement $
         compile (Loop i size   $
                       WriteArr (Var name arrTy) (Var i Int) elem)
  allocC <- allocateArray loc elemTy sizeC
  return (allocC $$ initC)
 where
   elemTy  = typeOf elem
   arrTy   = Arr elemTy

{- ToDo: verify that massageExpr will never produce MkArr's that
   are non-simple.  (conal) It does!
-- Complex size
compileMkArr name (MkArr size i elem) = compile $
  Let sizeName size (Let name (MkArr (Var sizeName Int) i elem)
 where
   sizeName = unsafePerformIO (newId' "size")
-}

-- Complex size
compileAssign loc name (MkArr size i elem) = 
 compileAssign loc name $
    Let sizeName size (MkArr (Var sizeName Int) i elem)
 where
   sizeName = unsafePerformIO (newId' "size")

compileAssign loc name (NewArr size elemTy) = do
  sizeC  <- compile size
  allocArray loc size sizeC elemTy (return ())
  allocateArray loc elemTy sizeC

compileAssign loc name (Perform e) = do
  compileAssign loc name e

compileAssign loc name (Bind a (Lambda v ty b)) = do
  aCode <-
	if v `occFreeIn` b then do
  	   loc   <- newLocal' v ty
	   compileAssign (ppLocal loc) v a
	 else
	   compile a	
  bCode <- compileAssign loc name b
  return (aCode <> semi $$ bCode <> semi)

compileAssign loc name (NewRef init) = do
  initC <- compile init
  return (loc <+> equals <+> initC <> semi)

compileAssign loc name e = do
  eC <- compile e
  return (loc <+> equals <+> eC <> semi)


joinIf :: CCode -> CCode -> CCode -> CgM CCode
joinIf condC thenC elseC = do
  flg   <- atStatementLevel
  if flg then
     return $
       text "if" <+> parens condC <+> char '{' $$
       nest 3 thenC $$
       char '}' <+> text "else" <+> char '{' $$
       nest 3 elseC $$
       char '}'
   else do
      --
      -- Expression level ifs aren't fully supported,
      -- since we require all occurrences of MkArr to
      -- be floated to the level of statements. So, if
      -- MkArr occurs inside the branches of an If, we're
      -- stuck since we cannot safely float the MkArr construction
      -- out of the If. 
      -- 
     return $
       parens $
         condC <+> char '?' <+> 
	    thenC <+> char ':' <+> elseC


{-
 Perform the rewrite f (Let binds e) ==> Let binds (f e).

 This stuff is (unfortunately) needed in the compilation of 
 MkArrs.
-}
floatBeyondLet :: (Exp -> Exp)
	       -> Exp
	       -> Exp
floatBeyondLet f (Let i rhs e) = Let i rhs (floatBeyondLet f e)
floatBeyondLet f e = f e


-- PP utils

ppBinder :: (Local,CCode) -> CCode
ppBinder ((n,_),e) = ppVar n <+> equals  <+> e


ppLocal :: Local -> CCode
ppLocal (n,_) = ppVar n

ppImportVar prefix n = text ("impD_"++prefix ++ '_':show n)

-- 

newtype CgM a 
 = CgM ( CgEnv -> CgState -> (a, CgState) )

returnCgM a = CgM (\ _ st -> (a,st))

bindCgM (CgM a) f = 
  CgM (\ env st ->
     case a env st of
        (v, st1) ->
	    let
	     (CgM b) = f v
	    in
	    b env st1)

type Imports = [(String,String,Int)]

data CgState =
  CgState
   { avails      :: AvailLocals
   , arrAvails   :: [((Type,Exp),Local)]
   , arrAllocs   :: [((Type,Exp),CCode)]
   , imports     :: (Int, Imports)
   , externs     :: [BoundVar]
   , boundLocals :: [BoundLocal]
   }

data CgEnv
  = CgEnv
      { params      :: [Param]
      , funName     :: String
      , opMap       :: (Op -> String)
      , atStmt      :: Bool  -- True => currently emitting C statement.
      , heapAlloc   :: Bool  -- True => allocate arrays on the heap.
      }

instance Monad CgM where
  return = returnCgM
  (>>=)  = bindCgM

runCgM :: CgM a     		-- code generating action
       -> [BoundVar]  		-- bound global variables.
       -> (a, Imports)
runCgM (CgM a) globals = 
  case (a (CgEnv [] "" (\ _ -> "") True False)
          (CgState ([],0) [] [] (0,[]) globals [])) of
    (v, CgState _ _ _ (_,imps) _ _) -> 
    	  (v, imps)

runCg  :: CgM a     		-- code generating action
       -> [Param]   		-- bound parameters
       -> [BoundVar]  		-- bound global variables.
       -> String
       -> (Op -> String)
       -> CgM (a, [Local])
runCg (CgM a) params globals nm opm = CgM $ \ env st -> 
  case (a (CgEnv params nm opm True False) st{boundLocals=[],externs=globals}) of
    (v, st1@CgState{avails=(ls,n)}) -> 
        ((v,ls),st1{boundLocals=[],avails=([],n)})

getEnv :: CgM CgEnv
getEnv = CgM (\ env st -> (env, st))

getLocalEnv :: CgM [BoundLocal]
getLocalEnv = CgM (\ env st -> (boundLocals st, st))

getExterns :: CgM [BoundVar]
getExterns = CgM (\ env st -> (externs st, st))

getParams :: CgM [Param]
getParams = CgM (\ env st -> (params env, st))

getAvailLocals :: CgM AvailLocals
getAvailLocals = CgM ( \ _ st -> (avails st, st))

asExpression :: CgM a -> CgM a
asExpression (CgM a) = CgM (\ env st -> a env{atStmt=False} st)

asStatement :: CgM a -> CgM a
asStatement (CgM a) = CgM (\ env st -> a env{atStmt=True} st)

atStatementLevel :: CgM Bool
atStatementLevel = CgM (\ env st -> (atStmt env, st))

allocHeap :: CgM a -> CgM a
allocHeap (CgM a) = CgM (\ env st -> a env{heapAlloc=True} st)

allocateOnHeap :: CgM Bool
allocateOnHeap = CgM (\ env st -> (heapAlloc env, st))

-- map between the name of an Exp var name (that isn't bound by a param) and
-- its corresponding C identifier.
lookupLocal :: Id -> CgM CCode
lookupLocal nm = do
  locals <- getLocalEnv
  exts   <- getExterns
  case (lookup nm exts) of
    Just (v,_) -> return (text v)
    _ -> 
     case  find ((== nm) . fst) locals of
        Just (_,(m,_)) -> return (ppVar m)
        Nothing    -> 
                  error ("Failed to find " ++ show nm ++
                         ".  Let-bound: " ++
                           showSep showBoundLocal locals
			)

newImport :: String -> CgM (String,Int)
newImport nm = 
  CgM (\ env st -> 
  	let (n, is) = imports st
	    pref    = funName env
	in
  	case find (\ (x,_,_) -> x == nm) is of
	  Nothing       -> ((pref, n), st{imports=(n+1,(nm,pref, n):is)})
	  Just (_,_,x)  -> ((pref, x), st))

newLocal' :: String -> Type -> CgM Local
newLocal' nm ty  =
  CgM (\ env st -> 
  	  let
           (loc, (ls,n)) = makeLocal ty (avails st)
	   orig_locals=boundLocals st
	   st1 = st{avails=(loc:ls,n),boundLocals=(nm,loc):orig_locals}
	  in
	  (loc, st1))

arrayScope :: CgM a -> CgM a
arrayScope (CgM x) = 
  CgM (\ env st ->
           let
	    as1 = arrAllocs st
	    st1 = st{arrAllocs=[]}
	   in
	   case x env st1 of
	     (v, st2) ->
	        let
	         as  = arrAllocs st1
--		 st3 = st2{arrAllocs=as1, arrAvails=as ++ (arrAvails st1)}
		in
		(v, st2))

allocArray :: CCode -> Exp -> CCode -> Type -> CgM a -> CgM (a, CCode)
allocArray loc size sizeC elemTy (CgM x) = 
  CgM (\ env st -> 
          let
	   arrs     = arrAvails st
	   allocs   = arrAllocs st
	   res      = lookup (elemTy,size) arrs
	   (Just l) = res

	   arrs' = arrs
{- ToDo: re-enable the reuse of array locals.
	    | not (isJust res) = arrs
	    | otherwise        = filter (\ x -> l /= snd x) arrs
-}    
	   new_allocs
--	    | not (isJust res) = ((elemTy, size), loc):allocs
	    | otherwise        = allocs

	   st1 = st{arrAvails = arrs'}
	   
	   code v
--	    | isJust res = return (v, ppLocal l)
	    | otherwise  = do
	        c <- allocateArray loc elemTy sizeC
		return (v,c)
	  in
	  case x env st1 of
	    (v, st2) ->
	       let
	        st3 = st2{arrAllocs=new_allocs}
		 -- tiresome - have to use a monadic prim.
		 -- from within another prim.
		(CgM x) = code v
	       in
	       x env st3)

allocateArray loc elemTy sizeC = do
   flg <- allocateOnHeap
   if flg then do
       return 
         (text "ALLOC_HEAP_ARRAY" <+>
	     ppTuple [ loc
	             , text (showCType elemTy ++ "*")
		     , text (if isArrayTy elemTy then "1" else "0")
		     , sizeC <+> char '*' <+> ppInt (sizeofType elemTy)
		     ] <> semi)
    else do
     return $
      text "ALLOC_ARRAY" <> 
      ppTuple [ loc
             , text (showCType elemTy ++ "*")
	     , sizeC <+> char '*' <+> ppInt (sizeofType elemTy)
	     ] <> semi

newLocal :: String -> Type -> (Local -> CgM a) -> CgM (a, Local)
newLocal nm ty m  = 
  CgM (\ env st -> 
  	  let
           (loc, avails') = makeLocal ty (avails st)
	   orig_locals=boundLocals st
	   st1 = st{avails=avails',boundLocals=(nm,loc):orig_locals}
	   (CgM x) = m loc
	  in
          case 	x env st1 of
            (v, st2) -> 
                let
                 (ls,n) = avails st2
                in
                ((v, loc), st2{avails=(loc:ls,n),boundLocals=orig_locals}))

-- Note: if you extend this list, be sure also to extend the lists in DExpUtils.hs

showCOp op =
  case op of
    AddOp        -> "+"
    SubOp        -> "-"
    NegateOp     -> " -"                  -- space to prevent -- when not optimizing
    MulOp        -> "*"
    RecipOp      -> "rcp" -- F32vec4 op
    DivideOp     -> "/"
    PowOp        -> "pow"
    LogBaseOp    -> "pan_logBase"
    SinOp        -> "sin"
    CosOp        -> "cos"
    Atan2Op      -> "atan2"
    SqrtOp       -> "sqrt"
    AbsOp        -> "pan_abs"
    I2fOp        -> "(float)"
    F2iOp        -> "(int)"
    B2iOp        -> "(int)"
    ModOp        -> "%"
    DivOp        -> "/"
    EqualOp      -> "=="
    LessThanOp   -> "<"
    MinOp        -> "pan_min"
    MaxOp        -> "pan_max"
    AndOp        -> "&&"
    OrOp         -> "||"
    NotOp        -> "!"
    LShiftOp     -> "<<"
    RShiftOp     -> ">>"
    PackF32Op    -> "F32vec4" -- C++ class constructor
    SumF32Op     -> "add_horizontal"
    MkTupleOp    -> "pan_tuple"
    TupleEltOp i -> "pan_elt" ++ show i
    BitAndOp     -> "&"
    BitOrOp      -> "|"
    BitNotOp     -> "~"
    Deref32Op    -> "*(int*)"
    SetIntOp     -> "=(int)"
    Deref8Op     -> "*(unsigned char*)"
    _            -> error ("compile: I don't yet handle " ++ show op)


showURTOp op =
  case op of
    AddOp      -> "+"
    SubOp      -> "-"
    NegateOp   -> "-"
    MulOp      -> "*"
    DivideOp   -> "/"
    PowOp      -> "Math.Pow"
    SinOp      -> "Math.Sin"
    CosOp      -> "Math.Cos"
    Atan2Op    -> "Math.Atan2"
    SqrtOp     -> "Math.Sqrt"
    I2fOp      -> "(double)"
    F2iOp      -> "(int)"
    B2iOp      -> ""
    ModOp      -> "%"
    DivOp      -> "/"
    EqualOp    -> "=="
    LessThanOp -> "<"
    AndOp      -> "&&"
    OrOp       -> "||"
    NotOp      -> "!"
    LShiftOp   -> "<<"
    RShiftOp   -> ">>"
    BitAndOp   -> "&"
    BitOrOp    -> "|"
    BitNotOp   -> "~"
    Deref32Op  -> "*(int*)"
    SetIntOp   -> "="
    Deref8Op   -> "*(unsigned char*)"
    _ -> error ("compile: I don't yet handle " ++ show op)


{-
 In order to simplify the translation of Exps to C, we massage
 the final Exp value first. The details of why this needs to be
 done is outlined in doc/c-translation.txt
-}

-- to highlight that we're mapping from unadulterated Exps
-- to a constrained form, introduce some type synonyms:
type CStmt = Exp 
type CExp = Exp 
 -- ToDo: state what the constraints are.

type NameSupply = [String] -- long list of labels.

nextVar :: NameSupply -> (String, NameSupply)
nextVar (x:xs) = (x,xs)

massageExpr :: Exp -> CStmt
massageExpr e = fst (massageExp supply e)
  where
   supply = map (\ x -> 'X':show x) [0..]

massageExp :: NameSupply -> Exp -> (CStmt, NameSupply)
massageExp supply exp =
  case exp of
    Let i rhs e        -> massageBind supply exp
    Loop i bound body  -> liftExp supply bound $ \ supply bound' -> 
    			  let
			   (body', supply1) = massageExp supply body
			  in
			  (Loop i bound' body', supply1)
    If c e1 e2         -> liftExp supply c $ \ supply1 c' -> 
    			  let
			   (e1', supply2) = massageExp supply1 e1
			   (e2', supply3) = massageExp supply2 e2
			  in
			  (If c' e1' e2', supply3)
    Bind a b           -> liftExp supply exp $ flip (,)
    MkArr size i elExp -> liftExp supply exp $ flip (,)
    NewArr size ty     -> liftExp supply exp $ flip (,)
    Perform e          -> liftExp supply exp $ flip (,)
    SetInt ptr val     -> liftExp supply exp $ flip (,)
--    NewRef v filler    -> liftExp supply exp $ flip (,)
--    UpdateRef ref newValF -> liftExp supply exp $ flip (,)
{-
    	liftExp supply ref $ \ supply1 ref' ->
	let
	 (newValF', supply2) = massageExp supply1 newValF
	in
	(UpdateRef ref' newValF', supply2)
-}
     -- assumption: every use of Lambda occurs within a
     -- construct that has been lifted up to the statement-level
     -- (NewRef, UpdateRef)
    Lambda v ty body ->
	let
	 (body', supply1) = massageExp supply body
	in
	(Lambda v ty body', supply1)
    x 
     | isSimple  x -> (x, supply)
     | otherwise   -> 
          let
	   (op, args) = toHnf exp
	  in
	  liftExps supply args (\ args' -> fromHnf (op, args'))

massageBind :: NameSupply 
	    -> Exp
	    -> (CStmt, NameSupply)
massageBind s e = liftExp s e $ \ supply e' -> (e', supply)

liftExps :: NameSupply
	 -> [Exp]
	 -> ([CExp] -> CStmt)
	 -> (CStmt, NameSupply)
liftExps supply args cont = 
    let
	-- process the arguments Exps right-to-left in order
	-- to have the bindings come out left-to-right without
	-- reversing.
     (bss, args', supply1) = 
	foldr (\ a (bss, as, supply) -> 
		 let
		  (bs, a', supply1) = liftExp' supply a
		 in
		 (bs:bss, a':as, supply1))
	      ([],[],supply)
      	      args
    in
    (foldr mkBind (cont args') (concat bss), supply1)

safeToUse MkArr{}  = False
safeToUse NewRef{} = False
safeToUse NewArr{} = False
safeToUse _	   = True

liftExp :: NameSupply 
	-> Exp
	-> (NameSupply -> CExp -> (CStmt, NameSupply))
	-> (CStmt, NameSupply)
liftExp supply e cont = 
	case (liftExp' supply e) of
	  (bs, e, supply') -> 
	  	let
		 (e', supply1) = cont supply' e
		in
		(foldr mkBind e' bs, supply1)

-- Rewrite (Let x rhs x) to rhs, as this form could
-- be introduced by liftExp'
mkBind (i,rhs) (Var v _) 
   | v == i && safeToUse rhs = rhs
mkBind (i,e1@(Var v _)) e2 = substVar i e1 e2
mkBind (i,rhs) e = Let i rhs e

liftExp' supply e =
   case e of
      Let i rhs e1 -> (bs1 ++ (i,rhs'):bs2, e1', supply2)
    	where
         (bs1, rhs', supply1) = liftExp' supply rhs
         (bs2, e1',  supply2) = liftExp' supply1 e1

      Bind a (Let x rhs e) -> liftExp' supply (Let x rhs (Bind a e))
      Bind a b     -> (bs1, Bind a' b', supply2)
       where
	 (bs1, a', supply1) = liftExp' supply a
	 (b', supply2) = massageExp supply1 b

	 
      NewArr sz ty -> (bs1 ++ [(v,NewArr sz' ty)], Var v (typeOf e), supply2)
       where
         (v, supply2) = nextVar supply1
	 (bs1, sz', supply1) = liftExp' supply sz
	 
      Perform a    -> (bs1 ++ [(v,Perform a')], Var v (typeOf e), supply2)
       where
         (bs1, a', supply1) = liftExp' supply a
         (v, supply2) = nextVar supply1
	 
      Loop i bound body -> (bs1, Loop i bound' body', supply2)
    	where
         (bs1, bound', supply1) = liftExp' supply bound
         (body', supply2) = massageExp supply1 body

      MkArr size i valAt -> (bs1 ++ [(v, MkArr size' i valAt')], Var v (typeOf e), supply3)
    	where
	 (v,supply1) = nextVar supply
         (bs1, size', supply2) = liftExp' supply1 size
         (valAt', supply3) = massageExp supply2 valAt

      If c e1 e2 -> (bs1 ++ [(v, If c' e1' e2')], Var v (typeOf e), supply4)
    	where
	 (v,supply1)        = nextVar supply
         (bs1, c', supply2) = liftExp' supply1 c
         (e1', supply3) = massageExp supply2 e1
         (e2', supply4) = massageExp supply3 e2

{-
      NewRef init filler -> ( bs1 ++ [(v, NewRef init' filler')]
      			    , Var v (typeOf e)
   			    , supply3
			    )
        where
	 (v, supply1) = nextVar supply
	 (bs1, init', supply2) = liftExp' supply1 init
	 (filler', supply3)    = massageExp supply2 filler
      UpdateRef ref newValF -> (bs1, UpdateRef ref' newValF', supply2)
        where
	 (bs1, ref', supply1) = liftExp' supply ref
	 (newValF', supply2) = massageExp supply1 newValF
-}

      Lambda v ty body -> ([], Lambda v ty body', supply1)
        where
	 (body', supply1) = massageExp supply body

      x
        | isSimple x -> ([], x, supply)
	| otherwise  ->
	     let
	      (op, args) = toHnf x
	      
		-- process the arguments Exps left-to-right
	      (bss, args', supply1) = 
	      	foldr (\ a (bss, as, supply) -> 
			 let
			  (bs, a', supply1) = liftExp' supply a
			 in
			 (bs:bss, a':as, supply1))
		      ([],[],supply)
	      	      args
		      
	      hasLet Let{} = True
	      hasLet _     = False
	     in
	     if any hasLet args' then error "the impossible happened in CCompile.liftExp'" else
	     (concat bss, fromHnf (op,args'), supply1)

