module Image where

import Exp
import PointE
import TransformE
import Color
-- import Arrays

infixr 3  `intersectR`, `diffR`
infixr 2  `unionR`


-- Static, spatially continuous image over arbitrary "pixel" type.
type Image c = PointE -> c

type ImageC  = Image ColorE
type ImageCR = Image BGR24E

-- Animated image
type ImageB c = TimeE -> Image c
type ImageBC  = ImageB ColorE
type ImageBCR = ImageB BGR24E

-- Filters, i.e., image to image functions
type Filter c = Image c -> Image c
type FilterC  = Filter ColorE

type FilterB c = Image c -> ImageB c
type FilterBC  = FilterB ColorE

type HyperFilter c = Filter c -> Filter c



empty :: p -> ColorE
empty = const invisible

cond :: Syntactic b => (a -> BoolE) -> (a -> b) -> (a -> b) -> (a -> b)
cond = lift3 ifE

over :: ImageC -> ImageC -> ImageC
over = lift2 overC

overs :: [ImageC] -> ImageC
overs = foldr over empty

fade :: Image FractionE -> ImageC -> ImageC
fade = lift2 fadeC

negI :: ImageC -> ImageC
negI = lift1 negC

lerpI :: Image FractionE -> ImageC -> ImageC -> ImageC
lerpI = lift3 lerpC

lightenI, darkenI :: Image FractionE -> FilterC
lightenI = lift2 lighten
darkenI  = lift2 darken

alphaI :: Image ColorE -> Image FractionE
alphaI = lift1 colorA

-- some monochromatic images

redI, orangeI, greenI, yellowI, blueI, whiteI, blackI :: ImageC

redI    = const red
orangeI = const orange
greenI  = const green
yellowI = const yellow
blueI   = const blue
whiteI  = const white
blackI  = const black

greyI :: FractionE -> ImageC
greyI x = const (grey x)

-- Turn a transform into a filter.  Note inversion.
imXf :: TransformE -> Filter c
imXf xf im = im . xf


-- Declarations like the following enable better type error messages from Hugs

translate, scale,tile        :: VectorE -> Filter c
rotate, uscale, swirl, shear :: FloatE  -> Filter c
cwave                        :: (FloatE -> FloatE) -> FloatE -> Filter c

-- Convenience functions.  Invert and compose
translate (dx,dy) = imXf $ translateP (-dx, -dy)
scale     (sx,sy) = imXf $ scaleP     (1/sx, 1/sy)
rotate    ang     = imXf $ rotateP    (-ang)
shear     dxdy    = imXf $ shearP (-dxdy)
tile      size    = imXf $ tileP    size
swirl     r       = imXf $ swirlP  r
cwave     f r     = imXf $ cwaveP   f r

-- Uniform scaling
uscale s = scale (s,s)

-- Horizontal, vertical reversal
hflip, vflip :: Filter c
hflip = scale (-1,1)
vflip = scale (1,-1)

bilerpC :: ColorE -> ColorE -> ColorE -> ColorE -> ImageC
-- Within a unit square.  Would need scaling if used on its own.
bilerpC ll lr ul ur = \ (dx,dy) ->
  lerpC dy (lerpC dx ll lr) (lerpC dx ul ur)

---- Regions

type Region  = Image BoolE
-- type RegionE = Region -- for better name consistency

-- Visualize a region as black & transparent (could use b&w)
regionToImageC :: Region -> ImageC
regionToImageC reg = crop reg blackI


intersectR, unionR, xorR, diffR :: Region -> Region -> Region
complementR                     :: Region -> Region

intersectR  = lift2 (&&*)
unionR      = lift2 (||*)
xorR        = lift2 (/=*)
complementR = lift1 notE

diffR r r' = r `intersectR` complementR r'

-- Universal and empty regions
universeR, emptyR :: Region
universeR = const true
emptyR    = const false

-- Half-planes.  "x positive", "y positive"
xPos, yPos :: Region
xPos = \ (x,y) -> x >* 0
yPos = \ (x,y) -> y >* 0


-- crop :: Region -> ImageC -> ImageC
crop :: (p -> BoolE) -> (p -> ColorE) -> (p -> ColorE)
crop reg im = cond reg im empty

-- Boolean checkerboard.  Square size is one, so you'll want to scale
checker :: Region
checker = \ (x,y) -> evenE (floorE x + floorE y)

udisk :: Region
udisk = \ p -> distOsq p <* 1

-- Annulus (ring) with unit outer radius and given inner radius
annulus :: FractionE -> Region
annulus inner p = inner*inner <=* r &&* r <* 1
 where
   r = distOsq p

-- This one is a bit more elegant, but fails to CSE the distOsq
-- computation.  Perhaps is would under GHC.
annulus' inner = udisk `diffR` disk inner

disk :: FloatE -> Region
disk r = uscale r udisk

diskIm :: FloatE -> ColorE -> ImageC
diskIm r c = crop (disk r) (const c)

box :: FloatE -> FloatE -> Region
box wid hgt (x,y) = abs x <=* wid/2 &&* abs y <=* hgt/2



withinI :: FloatE -> ImageC -> ImageC -> Region
withinI epsilon = lift2 (withinC epsilon)

chromaReg :: FloatE -> ColorE -> ImageC -> Region
chromaReg epsilon bg im = complementR (match invisible `unionR` match bg)
 where
   match col = withinI epsilon (const col) im

chromaKey :: FloatE -> ColorE -> FilterC
chromaKey epsilon bg im = crop (chromaReg epsilon bg im) im


-- Perform an image transformation "about" a point.  Like xfAbout.
about :: PointE -> HyperFilter c
about (x,y) imf = translate (x,y) . imf . translate (-x,-y)

-- Similarly for scaling & turning
atScale           :: VectorE -> HyperFilter c
atUScale,atRotate :: FloatE  -> HyperFilter c

atScale v@(sx,sy) xf = scale v . xf . scale(1/sx,1/sy)
atUScale s           = atScale (s,s)
atRotate ang xf      = rotate ang . xf . rotate (-ang)

-- Checker-board of given square size and colors
checkerBoard :: Syntactic c => FloatE -> Image c -> Image c -> Image c
checkerBoard sqSize = cond (uscale sqSize checker)


rippleRadius :: IntE -> FloatE -> Filter c
rippleRadius n s = imXf $ rippleRadiusP n (-s)

-- Pinwheel region.  Swirls rippled image
pinwheel :: FloatE -> IntE -> FloatE -> Filter c
pinwheel r n s = swirl r . rippleRadius n s

-- Radially alternating image (boolean pie slices)
radReg :: IntE -> Region
radReg arms = test . snd . toPolar
  where test a = evenE (floorE (a * i2f arms / pi))

-- Crop an image
radCrop :: IntE -> FilterC
radCrop = crop . radReg

-- Alternative:
wedges :: IntE -> FloatE -> Region
wedges arms frac =
  rotate (pi/2 + pi * (1-frac) / i2f arms) $   -- for symmetry about Y axis
  (f . snd . toPolar)
 where
   f a = fracPart (a' * i2f arms / (2*pi)) <=* frac
    where
      -- atan2/toPolar yields angle in [-pi,pi]
      a' = a + ifE (a<*0) (2*pi) 0

-- Inner swirls.  Acts only within a given radius and swirls more toward
-- center.

iswirl :: BoolE -> FloatE -> FloatE -> Filter c
iswirl inside rad sw = imXf $ iswirling inside rad (-sw)

radInvert :: Filter c
radInvert im =  im. radInvertP


-- How light is the image?  Average of r,g,b components.  What's the right
-- definition or name?
lightnessI, darknessI :: ImageC -> Image FractionE
lightnessI = lift1 lightness
darknessI  = lift1 darkness

-- Make lighter parts more transparent
transpLightI :: ImageC -> ImageC
transpLightI im = fade (darknessI im) im

-- Test whether each color is legitimate
legitI :: ImageC -> Region
legitI = lift1 legitC


---- Generalized tiling

type ImageBXf c = ImageB c -> ImageB c

gTile :: (DoubleE, DoubleE) -> ((IntE,IntE) -> DoubleE) -> ImageBXf c

gTile (w,h) dt imb t (x,y) = imb (t + dt (tx,ty)) (dx,dy)
 where
   (tx,dx) = x `divM` w
   (ty,dy) = y `divM` h
   z `divM` lim = (tnum, z - i2f tnum * lim)
    where l2   = lim/2
          z'   = z + l2
          tnum = floorE (z'/lim)

